from importlib import import_module
from django.apps import apps
from django.core.exceptions import PermissionDenied

appConfig = apps.get_app_config('document_catalogue')

permissions = import_module(appConfig.settings.PERMISSIONS)


def permission_required(permission_fn):
    """
        Constructs a CBV decorator that checks permission_fn(view.request.user, view.kwargs) before calling
          view.dispatch to test if request.user has a given (object) permission.
        Usage:  @permission_required("permission_function_name")  class MyViewClass: ...
    """
    def decorator(view_class):
        _dispatch = view_class.dispatch

        def dispatch(self, request, *args, **kwargs):
            if not permission_fn(request.user, **self.kwargs):
                raise PermissionDenied()
            return _dispatch(self, request, *args, **kwargs)

        view_class.dispatch = dispatch
        return view_class

    return decorator



from django.conf import settings
from django.apps import apps
from django.urls import path, include
from django.conf.urls.static import static

appConfig = apps.get_app_config('document_catalogue')

urlpatterns = [

    path('documents/', include('document_catalogue.urls')),

]

if appConfig.USE_PRIVATE_FILES:
    urlpatterns += [
        path('private-media/', include('private_storage.urls')),

    ]

urlpatterns += static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)



from collections import namedtuple
from django.apps import AppConfig
import django.conf

def Map(type_name, **kwargs):
    # Returns a simple map of kwargs object
    mapType = namedtuple(type_name, kwargs.keys(),)
    return mapType(**kwargs)

class BaseCatalogueConfig(AppConfig):
    name = 'document_catalogue'
    verbose_name = 'Document Catalogue'

    # Default settings for the document_catalogue app.
    settings = Map(
        'default_settings',
        # Fine scale permissions (default permissions use DOCUMENT_CATALOGUE_LOGIN_REQUIRED, but that setting may be disabled by custom permissions)
        # Value is a dotted path to a permissions module or object with the required permissions functions - see permissions.py
        PERMISSIONS = getattr(django.conf.settings,
                              'DOCUMENT_CATALOGUE_PERMISSIONS', 'document_catalogue.permissions'),

        # Enable document editing URL's and the Ajax document API
        # When disabled, documents are managed via the django.admin
        # When enabled, be sure to carefully consider the permissions on these views!
        ENABLE_EDIT_URLS = getattr(django.conf.settings,
                                   'DOCUMENT_CATALOGUE_ENABLE_EDIT_URLS', False),

        # Root directory for file uploads to the document catalogue
        # If private-files is used, media_root will be a sub-directory of PRIVATE_STORAGE_ROOT
        MEDIA_ROOT = getattr(django.conf.settings,
                             'DOCUMENT_CATALOGUE_MEDIA_ROOT', 'documents/'),

        # Upload file settings - constrain the file content_types and max. filesize for document file uploads.
        # None to allow any file type - eliminates dependency on python-magic and libmagic, which are included only if a whitelist is supplied.
        #   e.g.  DOCUMENT_CATALOGUE_CONTENT_TYPE_WHITELIST = ('application/pdf', 'image/png', 'image/jpg')
        CONTENT_TYPE_WHITELIST = getattr(django.conf.settings,
                                         'DOCUMENT_CATALOGUE_CONTENT_TYPE_WHITELIST', None),
        # max. size in bytes, None for no limit
        MAX_FILESIZE = getattr(django.conf.settings,
                               'DOCUMENT_CATALOGUE_MAX_FILESIZE', 10 * 1024 * 1024),

        # Plugin Classes used to inject behaviours into standard document list views
        # Define plugins by extending the ABC: document_catalogue.plugins.AbstractViewPlugin
        # Inject them by customizing this setting
        LIST_VIEW_PLUGINS = getattr(django.conf.settings,
                                    'DOCUMENT_CATALOGUE_LIST_VIEW_PLUGINS',
                                    ('document_catalogue.plugins.SessionOrderedViewPlugin',)),
   )


class PrivateCatalogueConfig(BaseCatalogueConfig):
    verbose_name = 'Private Catalogue'

    USE_PRIVATE_FILES = True
    LOGIN_REQUIRED_DEFAULT = True

    settings = Map(
        "private_settings",
        USE_PRIVATE_FILES = True,
        LOGIN_REQUIRED = getattr(django.conf.settings,
                                 'DOCUMENT_CATALOGUE_LOGIN_REQUIRED', True),
        **BaseCatalogueConfig.settings._asdict()
    )



class PublicCatalogueConfig(BaseCatalogueConfig):
    verbose_name = 'Public Catalogue'

    USE_PRIVATE_FILES = False
    LOGIN_REQUIRED_DEFAULT = False
    settings = Map(
        "public_settings",
        USE_PRIVATE_FILES=False,
        # IMPORTANT: this only restricts access to the Catalogue views - anyone with a file's URL can download it!
        LOGIN_REQUIRED = getattr(django.conf.settings,
                                 'DOCUMENT_CATALOGUE_LOGIN_REQUIRED', False),
        **BaseCatalogueConfig.settings._asdict()
    )



""" Settings for PrivateCatalogueConfig tests """
from .base import *

INSTALLED_APPS += [
    'document_catalogue.apps.PrivateCatalogueConfig',
    'private_storage',
]

# django-private-files  (private document downloads)
PRIVATE_STORAGE_ROOT = os.path.join(BASE_DIR, "private-media/")
PRIVATE_STORAGE_AUTH_FUNCTION = 'private_storage.permissions.allow_authenticated'



#-*- coding: utf-8 -*-

# django-sample-app - A Django app with setup, unittests, docs and demo
# Copyright (C) 2013,  Daniel Rus Morales

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import unittest


def setup_django_settings():
    os.chdir(os.path.join(os.path.dirname(__file__), ".."))
    sys.path.insert(0, os.getcwd())
    os.environ["DJANGO_SETTINGS_MODULE"] = "document_catalogue.tests.settings.private"


def run_tests():
    if not os.environ.get("DJANGO_SETTINGS_MODULE", False):
        setup_django_settings()

    import django
    from django.conf import settings
    from django.test.utils import get_runner

    django.setup()
    runner = get_runner(settings,"django.test.runner.DiscoverRunner")
    test_suite = runner(verbosity=2, interactive=True, failfast=False)
    return test_suite.run_tests(["document_catalogue"])



from django.test import TestCase
from document_catalogue import plugins

class FrameworkTests(TestCase):
    """
        Test basic behaviours for Plugin Micro-Framework
    """
    class TestPluginManager(plugins.PluginManager):
        def __init__(self):
            self.counter = 0

        def count(self):
            self.counter += 1

    def test_plugin_manager(self):
        class Subclass(self.TestPluginManager):
            pass

        t = Subclass()
        plugins = ('a', 'b', 'c')
        t.add_plugins(plugins)
        self.assertEqual(t.plugins, list(plugins))

        t.apply_plugins(lambda plugin: t.count())
        self.assertEqual(t.counter, len(plugins))

    def test_plugin_manager_inheritance(self):
        class Subclass1(self.TestPluginManager):
            pass
        class Subclass2(self.TestPluginManager):
            pass
        class SubSubclass(Subclass1):
            pass

        t = Subclass1()
        plugins = ('a', 'b', 'c')
        t.add_plugins(plugins)

        s = Subclass2()
        more_plugins = ('d', 'e')
        s.add_plugins(more_plugins)

        s.apply_plugins(lambda plugin: s.count())
        self.assertEqual(s.counter, len(more_plugins))

        r = SubSubclass()
        r.add_plugins(more_plugins)
        self.assertEqual(set(t.plugins), set(plugins))
        self.assertEqual(set(s.plugins), set(more_plugins))
        self.assertEqual(set(r.plugins), set(plugins) | set(more_plugins))

        r.apply_plugins(lambda plugin: r.count())
        self.assertEqual(r.counter, len(plugins) + len(more_plugins))


    def test_register_plugins(self):
        @plugins.RegisterPlugins('w', 'x', 'y', 'z')
        class Subclass(self.TestPluginManager):
            pass

        s = Subclass()
        s.apply_plugins(lambda plugin: s.count())
        self.assertEqual(s.counter, 4)  # one for each plugin added above


class ViewPluginTests(TestCase):
    """
        Test behaviours for View Plugins
    """
    class TestViewPluginManager(plugins.ViewPluginManager):
        pass

    class TestViewPlugin(plugins.AbstractViewPlugin):
        def apply(self, request):
            request.count += 1

        def extend_qs(self, request, qs):
            return qs + 10

        def get_context(self, request):
            return {'plugin_context': 'Some Value'}

    def setUp(self):
        @plugins.RegisterPlugins(self.TestViewPlugin(), self.TestViewPlugin())
        class Subclass(self.TestViewPluginManager):
            pass

        self.plugin_manager = Subclass()

    def test_apply_plugins(self):
        request = lambda : None
        request.count = 0
        self.plugin_manager.apply_plugins(lambda plugin: plugin.apply(request))
        self.assertEqual(request.count, 2)

    def test_extend_qs(self):
        qs = 0
        qs = self.plugin_manager.plugins_extend_qs(None, qs)
        self.assertEqual(qs, 20)

    def test_get_context(self):
        ctx = self.plugin_manager.plugins_get_context(None)
        self.assertTrue('plugin_context' in ctx)


class OrderedViewPluginTestBase(TestCase):
    """
        Set up for OrderedView Plugin Tests
    """
    class TestViewPluginManager(plugins.ViewPluginManager):
        pass
    ORDER_KEY = 'the_order_key'
    ORDER_FIELD = 'date'
    ORDER_EXPERSSION = plugins.OrderedViewPlugin.ORDERING_EXPRESSION[ORDER_FIELD]
    PLUGIN = None

    def setUp(self):
        @plugins.RegisterPlugins(self.PLUGIN)
        class Subclass(self.TestViewPluginManager):
            pass

        self.plugin_manager = Subclass()

        self.request = lambda : None
        self.request.GET = {self.ORDER_KEY: self.ORDER_FIELD}
        self.request.session = {}

        class QS:
            def __init__(self):
                self.ordering = None
            def order_by(self, ordering):
                self.ordering = ordering
        self.qs = QS()


class OrderedViewPluginTests(OrderedViewPluginTestBase):
    """
        Test behaviours for OrderedView Plugins
    """
    PLUGIN = plugins.OrderedViewPlugin(OrderedViewPluginTestBase.ORDER_KEY)

    def test_extend_qs(self):
        qs = self.plugin_manager.plugins_extend_qs(self.request, self.qs)
        self.assertEqual(self.qs.ordering, self.ORDER_EXPERSSION)

    def test_get_context(self):
        ctx = self.plugin_manager.plugins_get_context(self.request)
        self.assertTrue(self.ORDER_KEY in ctx)
        self.assertEqual(ctx[self.ORDER_KEY], self.ORDER_FIELD)
        self.assertTrue(self.ORDER_KEY+'_choices' in ctx)


class SessionOrderedViewPluginTests(OrderedViewPluginTestBase):
    """
        Test behaviours for SessionOrderedView Plugins
    """
    PLUGIN = plugins.SessionOrderedViewPlugin(OrderedViewPluginTestBase.ORDER_KEY)

    def test_apply_plugins(self):
        self.plugin_manager.apply_plugins(lambda plugin: plugin.apply(self.request))
        self.assertEqual(self.request.session[self.ORDER_KEY], self.ORDER_FIELD)

    def test_extend_qs(self):
        self.plugin_manager.apply_plugins(lambda plugin: plugin.apply(self.request))
        qs = self.plugin_manager.plugins_extend_qs(self.request, self.qs)
        self.assertEqual(self.qs.ordering, self.ORDER_EXPERSSION)

    def test_get_context(self):
        self.plugin_manager.apply_plugins(lambda plugin: plugin.apply(self.request))
        ctx = self.plugin_manager.plugins_get_context(self.request)
        self.assertTrue(self.ORDER_KEY in ctx)
        self.assertEqual(ctx[self.ORDER_KEY], self.ORDER_FIELD)
        self.assertTrue(self.ORDER_KEY+'_choices' in ctx)



import os
from django.apps import apps
import django.conf
from django.urls import reverse
from django.db import models
import mptt.models

appConfig = apps.get_app_config('document_catalogue')

class DocumentCategory(mptt.models.MPTTModel):
    """
        A hierarchical category system for assets
    """
    name = models.CharField(max_length=64)
    slug = models.SlugField(unique=True)
    # sort_order = models.SmallIntegerField(default=1)
    description = models.TextField(blank=True)
    parent = mptt.models.TreeForeignKey('self', on_delete=models.CASCADE, null=True, blank=True,
                                        related_name='children')

    objects = mptt.models.TreeManager()

    class MPTTMeta:
        order_insertion_by = ['slug']

    class Meta:
        verbose_name_plural = 'Categories'
        ordering = ['slug']

    def __str__(self):
        return str(self.name)

    def get_absolute_url(self):
        return reverse('document_catalogue:category_list', kwargs={'slug': self.slug, })

    def has_children(self):
        """ Return true if this DocumentCategory has dependent objects lower in the hierarchy """
        return self.get_descendant_count() > 0

    def get_document_count(self):
        """ Return the total number of documents uploaded for this category """
        return self.document_set.count()


class DocumentQueryset(models.QuerySet):
    """ Custom query set for Document model """
    def published(self):
        return self.filter(is_published=True)

    def in_category(self, category_slug):
        return self.filter(category__slug=category_slug)


BaseDocumentManager = models.Manager.from_queryset(DocumentQueryset)


class DocumentManager(BaseDocumentManager):
    """ Manager for Document model with all methods from DocumentQuerySet. """
    def get_queryset(self):
        return super().get_queryset().select_related('category')


class PublishedDocumentManager(DocumentManager):
    """ Published documents only """
    def get_queryset(self):
        return super().get_queryset().published()


def document_upload_path_callback(instance, filename):
    """ Dynamic upload path based on file instance """
    path = "%s%s/%s" % (appConfig.settings.MEDIA_ROOT, instance.category.slug, filename)
    return path


class Document(models.Model):
    """
        A file and the meta-data describing that file in the catalogue.
    """
    category = models.ForeignKey(DocumentCategory, on_delete=models.CASCADE)

    sort_order = models.PositiveSmallIntegerField(default=0, verbose_name='Order')

    user = models.ForeignKey(django.conf.settings.AUTH_USER_MODEL, on_delete=models.SET(1))

    creation_date = models.DateTimeField(auto_now_add=True, verbose_name='Created')

    update_date = models.DateTimeField(auto_now=True, verbose_name='Last Modified')

    title = models.CharField(max_length=512)

    description = models.TextField(null=True, blank=True)

    is_published = models.BooleanField(default=False)

    if appConfig.settings.USE_PRIVATE_FILES:
        from private_storage.fields import PrivateFileField
        file = PrivateFileField(
            upload_to=document_upload_path_callback,
            content_types=appConfig.settings.CONTENT_TYPE_WHITELIST,
            max_file_size=appConfig.settings.MAX_FILESIZE \
                if appConfig.settings.MAX_FILESIZE else None
        )
    else:
        from constrainedfilefield.fields import ConstrainedFileField
        file = ConstrainedFileField(
            max_length=200,
            upload_to=document_upload_path_callback,
            content_types=appConfig.settings.CONTENT_TYPE_WHITELIST,
            max_upload_size=appConfig.settings.MAX_FILESIZE * 1024 * 1024 \
                if appConfig.settings.MAX_FILESIZE else 0
        )
    objects = DocumentManager()
    published = PublishedDocumentManager()

    class Meta:
        ordering = ('sort_order', )

    def __str__(self):
        return self.title

    def get_absolute_url(self):
        return reverse('document_catalogue:document_detail', kwargs={'pk': self.pk, })

    def get_edit_url(self):
        return reverse('document_catalogue:document_edit', kwargs={'pk': self.pk, })

    def get_download_url(self):
        return reverse('document_catalogue:document_download', kwargs={'pk': self.pk, })

    def get_filetype(self):
        name, extension = os.path.splitext(self.file.name)
        return extension[1:]

    def filename(self):
        return os.path.basename(self.file.name)



"""
Django settings for testing document_catalogue.
"""

from pathlib import Path
import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = '/tmp/document_catalogue_tests/'

# Ensure the base path exists
Path(BASE_DIR).mkdir(parents=True, exist_ok=True)

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'foobar'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'mptt',
    # 'document_catalogue.apps.xConfig  added by concrete settings files
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'document_catalogue.tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'demo/templates')]
        ,
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Media files (document uploads)
# https://docs.djangoproject.com/en/2.2/howto/static-files/#serving-uploaded-files-in-development

MEDIA_ROOT = os.path.join(BASE_DIR, "media/")
MEDIA_URL = '/media/'

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/

STATIC_URL = '/static/'

STATICFILES_DIRS = (
    os.path.join(BASE_DIR, "static/"),
    # '/path/to/some_other_static_media',
)

# Document Catalogue overrides
DOCUMENT_CATALOGUE_LOGIN_REQUIRED = True
DOCUMENT_CATALOGUE_ENABLE_EDIT_URLS = True
DOCUMENT_CATALOGUE_CONTENT_TYPE_WHITELIST = ('text/plain', )
DOCUMENT_CATALOGUE_MAX_FILESIZE = 1 * 1024 * 1024



"""
Functions used to control access to views (Permission Denied returned when false)
    and to enable/disable interactions in the templates.

Default permissions module can be swapped out in settings.
Defaults use django's built in content_type permissions for the Document model.

Each function takes the request use and the view's kwargs as arguments,
    returns True iff user has the required permission for the given object(s).
"""
from django.apps import apps
appConfig = apps.get_app_config('document_catalogue')


def user_can_view_document_catalogue(user, **kwargs):
    """ Return True iff given user is allowed to view the document catalogue """
    return not appConfig.settings.LOGIN_REQUIRED or user.is_authenticated


def user_can_download_document(user, **kwargs):
    """ Return True iff the given user is allowed to download documents"""
    return user_can_view_document_catalogue(user, **kwargs)


def user_can_edit_document(user, **kwargs):
    """ Return True iff the given user can edit documents in the catalogue """
    return user.is_staff or user.has_perm('document_catalogue.change_document')


def user_can_post_document(user, **kwargs):
    """ Return True iff the given user can upload new files to the catalouge """
    return user.is_staff or user.has_perm('document_catalogue.add_document')


def user_can_delete_document(user, **kwargs):
    """ Return True iff the given user can delete documents from the catalogue """
    return user.is_staff or user.has_perm('document_catalogue.delete_document')



from django import forms
from document_catalogue.fields import ClearableFileWidget
from . import models


class DocumentUploadForm(forms.ModelForm):

    class Meta:
        model = models.Document
        fields = ('file', )


class DocumentEditForm(forms.ModelForm):
    file = forms.FileField(widget=ClearableFileWidget())

    class Meta:
        model = models.Document
        fields = ('title', 'description', 'category', 'is_published', 'file')

        widgets = {
            'title'      : forms.TextInput(attrs={'placeholder': 'Document Title (defaults to file name)'}),
            'description': forms.Textarea(attrs={'rows': 2, 'placeholder': 'Optional Description'}),
        }


import sys, os, re

from setuptools import setup, Command, find_packages
from setuptools.command.test import test

class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""
    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        os.system('rm -vrf ./*.pyc ./*.egg-info')


def run_tests(*args):
    from document_catalogue.tests import run_tests
    errors = run_tests()
    if errors:
        sys.exit(1)
    else:
        sys.exit(0)

test.run_tests = run_tests

NAME = "django-document-catalogue"

# get version without importing
with open("document_catalogue/__init__.py", "rb") as f:
    VERSION = str(re.search('__version__ = "(.+?)"', f.read().decode()).group(1))

setup(
    name=NAME,
    version=VERSION,
    packages=find_packages(include=['document_catalogue', 'document_catalogue.*']),
    python_requires='>=3.5, <4',
    install_requires = [
        'Django>=2.2,<3.0',
        'django-mptt>=0.10.0',
        'django-private-storage>=2.2',
        'django-admin-sortable2',
        'django-constrainedfilefield>=3.2.0',
        'python-magic>=0.4.15',
        'setuptools-git',    # apparently needed to handle include_package_data from git repo?
    ],
    license="MIT",
    include_package_data=True,  # declarations in MANIFEST.in
    description=("Simple, light-weight, stand-alone, hierarchical document library as a reusable django app."),
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    author="powderflask",
    author_email="powderflask@gmail.com",
    maintainer="powderflask",
    maintainer_email="powderflask@gmail.com",
    url="https://github.com/powderflask/django_document_catalogue",
    download_url="https://github.com/powderflask/django_document_catalogue/archive/v{}.tar.gz".format(VERSION),
    project_urls={
        'ReadTheDocs'  : 'https://django-document-catalogue.readthedocs.io',
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Framework :: Django',
    ],
    cmdclass={
        'clean' : CleanCommand,
    },
    test_suite="dummy",
)




from importlib import import_module
from functools import partial
from itertools import groupby

from django.apps import apps
from django.utils.module_loading import import_string
from django.utils.functional import cached_property
from django.urls import reverse
from django.shortcuts import get_object_or_404
from django.http import Http404, HttpResponseForbidden
from django.template.loader import get_template
from django.db.models import Q, Prefetch, Count
from django.views import generic

from .models import Document, DocumentCategory
from .views_generic import AjaxOnlyViewMixin
from .decorators import permission_required
from . import forms, plugins

appConfig = apps.get_app_config('document_catalogue')

# import Plugin Permissions module
permissions = import_module(appConfig.settings.PERMISSIONS)

# import Plugin classes
list_view_plugin_classes = tuple(import_string(plugin) for plugin in appConfig.settings.LIST_VIEW_PLUGINS)


def get_permissions_context(view):
    """ Return a dictionary of permissions (partials that can be called with no arguments) """
    context = {}
    for name in dir(permissions):
        fn = getattr(permissions, name)
        if callable(fn):
            context[name] = partial(fn, view.request.user, **view.kwargs)
    return context


@permission_required(permissions.user_can_view_document_catalogue)
class CatalogueViewMixin(generic.base.ContextMixin, generic.View):
    """ Mixin for all Document Views """
    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx.update({
            'show_edit_links': True if appConfig.settings.ENABLE_EDIT_URLS else False,
            **get_permissions_context(self),
        })
        return ctx


class CategorySlugViewMixin:
    """ Mixin for views that take a category slug as a URL arg """
    @property
    def category_slug(self):
        return self.kwargs.get('slug', None)

    @cached_property
    def category(self):
        return get_object_or_404(DocumentCategory, slug=self.category_slug)


class DocumentPkMixin:
    """ Mixins for views that take a document pk as a URL arg """
    @property
    def document_pk(self):
        return self.kwargs.get('pk', None)

    @cached_property
    def document(self):
        try:
            return Document.published.get(pk=self.document_pk)
        except Document.DoesNotExist:
            raise Http404


class DocumentCatalogueListView(CatalogueViewMixin, generic.ListView):
    """ List all categories in the Catalogue """
    template_name = 'document_catalogue/categories_list.html'

    queryset = DocumentCategory.objects.all()


class CatgetoryContextViewMixin(generic.base.ContextMixin, CategorySlugViewMixin):
    """ Mixing for views that supply context about a category """
    # here to serve as a base class so consitent MRO can be created
    pass


class CategoryListViewMixin(CatgetoryContextViewMixin):
    """ Mixin for views that navigate categories or display a list of categories """
    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx.update({
            'category': self.category,
        })
        return ctx


class BaseDocumentListView(plugins.ViewPluginManager,
                           CatalogueViewMixin, CatgetoryContextViewMixin, generic.ListView):
    """
        Base class / mixin for views that list documents.
        Plugin architecture used to inject custom view logic.  See plugins.
    """
    queryset = Document.published.all()
    def dispatch(self, request, *args, **kwargs):
        """ Apply any plugins """
        self.apply_plugins(lambda plugin: plugin.apply(request))
        return super().dispatch(request, *args, **kwargs)

    def get_document_queryset(self):
        qs = super().get_queryset()
        if self.category_slug:
            qs = qs.filter(category__slug=self.category_slug)
        return self.plugins_extend_qs(self.request, qs)

    def get_queryset(self):
        return self.get_document_queryset()

    def get_context_data(self, **kwargs):
        plugin_ctx = self.plugins_get_context(self.request)
        return super().get_context_data(**plugin_ctx)


@plugins.RegisterPlugins(*(plugin() for plugin in list_view_plugin_classes))
class CategoryDocumentListView(CategoryListViewMixin, BaseDocumentListView):
    """ List all documents in a given category """
    template_name = 'document_catalogue/category_document_list.html'


class DocumentViewMixin(generic.base.ContextMixin, DocumentPkMixin):
    """ Mixins for views that display a document """
    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx.update({
            'document' : self.document,
            'category' : self.document.category,
        })
        return ctx

    def get_document_absolute_uri(self):
        return self.request.build_absolute_uri(self.document.get_download_url())


class DocumentDetailView(CatalogueViewMixin, DocumentViewMixin,  generic.DetailView):
    """ Display detailed information about a single document """
    template_name = 'document_catalogue/document_detail.html'
    model = Document


@permission_required(permissions.user_can_view_document_catalogue)
class DocumentDownloadView(DocumentPkMixin,  generic.RedirectView):
    """ Redirect to the file URL """

    def get_redirect_url(self, *args, **kwargs):
        """ Return the document's file download URL """
        return self.document.file.url


@permission_required(permissions.user_can_edit_document)
class DocumentEditView(CatalogueViewMixin, DocumentViewMixin, generic.UpdateView):
    """ Display detailed information about a single document """
    template_name = 'document_catalogue/document_edit.html'
    model = Document
    form_class = forms.DocumentEditForm


@permission_required(permissions.user_can_delete_document)
class DocumentDeleteView(CatalogueViewMixin, DocumentViewMixin, generic.DeleteView):
    """ Delete a single document """
    model = Document

    @property
    def document(self):
        return self.object

    def get_success_url(self):
        return reverse('document_catalogue:category_list', kwargs={'slug':self.document.category.slug})


class DocumentAjaxAPI(CatalogueViewMixin, CategorySlugViewMixin, DocumentPkMixin, AjaxOnlyViewMixin):
    """
        Async API for document actions
        Plays nice with document_catalogue.js and dropzone
    """
    def save_document(self):
        file = self.request.FILES['file']
        document = Document(
            user=self.request.user,
            title=file.name,
            category=self.category,
            is_published=True,
            file=file
        )
        document.save()
        return document

    def post(self, request, *args, **kwargs):
        if not permissions.user_can_post_document(request.user, **self.kwargs):
            return HttpResponseForbidden('Permission Denied')

        form_class = forms.DocumentUploadForm
        document_template = get_template('document_catalogue/include/documents.html')

        def get_form():
            return form_class(data=request.POST, files=request.FILES)

        # Use the Upload Form to validate the file (mime type and size)
        form = get_form()
        if form.is_valid():
            document = self.save_document()
            html = document_template.render({'document_list': (document, ), **get_permissions_context(self)}),

            return self.render_to_json_response({
                'success': True,
                'document_item': html,
            })
        else:  # Common error handling is completed by dropzone -- this is a hard-fail fallback.
            return HttpResponseForbidden('Invalid request: Form errors %s' % ', '.join(e.as_text() for e in form.errors.values()))

    def delete(self, request, *args, **kwargs):
        if not permissions.user_can_delete_document(request.user, **self.kwargs) :
            return HttpResponseForbidden('Permission Denied')
        if self.document:
            self.document.delete()
            json_context = {"success" : True}
            return self.render_to_json_response(json_context)
        else:
            return HttpResponseForbidden('Invalid request. Document NOT deleted.')

    def get(self, request, *args, **kwargs):
        """ Ajax Search for documents matching search term in ?q= request param """
        search_term = request.GET.get('q', None)

        def format_select2(document):
            return {
                'id'  : document.get_absolute_url(),
                'text': document.title}

        search_options = []
        # Format options as select2 data objects
        if search_term:  # retrieve search results, if a search_term is given
            filter = Q(title__icontains=search_term) | \
                     Q(category__name__icontains=search_term)
            docs = Document.published.filter( filter )

            search_options = [
                {'text': category.name, 'children': [format_select2(doc) for doc in group] }
                for category, group in groupby(docs, lambda d: d.category)
            ]

        else:  # Recently updated documents.
            recently_updated = Document.published.order_by('-update_date')[:10]
            if recently_updated:
                options = [format_select2(doc) for doc in recently_updated]
                search_options = [{'text': 'Recently Updated', 'children':  options},]

        return self.render_to_json_response({'options': search_options})
